"use strict";

(function(global){

    function traverseJSEnvironnement(visitors, blackList){
        var visited = [];
        var toVisit = [ {object: global, 
                         access: []} ];
        var inVisit, objInVisit, accessInVisit, 
            oneStepAccess; // path from context to objInVisit object
        var proto;
        var propNames;

        function isNotVisitedOrAboutTo(o){
            return visited.indexOf(o) === -1 && toVisit.every(function(e){return e.object !== o;});
        }

        blackList = blackList || [];
        function inBlackList(e){
            return blackList.indexOf(e) !== -1;
        }

        if(typeof visitors === "function")
            visitors = [visitors];
        if(!Array.isArray(visitors))
            throw new TypeError("The visitors argument is supposed to be a function or an array of functions. It is a "+typeof visitors);

        while(toVisit.length !== 0){
            inVisit = toVisit.pop();

            objInVisit = inVisit.object;
            accessInVisit = inVisit.access;

            // Trying to find other unvisited objects based on inVisit
            // [[Prototype]]
            proto = Object.getPrototypeOf(objInVisit);
            if(proto !== null && isNotVisitedOrAboutTo(proto))
            toVisit.push( {object: proto, 
                           access: accessInVisit.concat(traverseJSEnvironnement.ACCESS_PROTO)} );
              
            // Properties
            propNames = Object.getOwnPropertyNames(objInVisit);
            console.log('list', accessInVisit, propNames);
            propNames.forEach(function(n){
                                  var desc;
                                  var v;
                                  
                                  try{ // Firefox throws exceptions here on some host objects
                                      desc = Object.getOwnPropertyDescriptor(objInVisit, n);
                                  }
                                  catch(e){
                                      // Do nothing, just move on
                                      console.log('Error thrown while trying to get the property desciptor of property '+ n);
                                      return;
                                  }
                                  
                                  if('value' in desc){ // data property descriptor because the traverser should not be calling getters
                                      v = objInVisit[n];
                                      if(v !== null && (typeof v === 'object' || typeof v === 'function') && isNotVisitedOrAboutTo(v)){
                                          toVisit.push( {object: v, 
                                                         access: accessInVisit.concat(n)} );
                                      }
                                  }
                              });

            // Call the visitors on the elements
            if(!inBlackList(objInVisit)){
                console.log('traverser', accessInVisit);
                visitors.forEach(function(v){
                                     v.call(objInVisit, accessInVisit);
                                 });
            }
            visited.push(objInVisit);
                           
        }
    }

    // Add symbolic constants
    Object.defineProperty(traverseJSEnvironnement, 'ACCESS_PROTO', {value: Object.preventExtensions( Object.create(null) ),
                                                                    configurable: false,
                                                                    writable: false,
                                                                    enumerable: false});

    // TODO: document path format
    traverseJSEnvironnement.resolvePath = function(path){
                                              return path.reduce(function(prev, curr){
                                                                     return (curr === traverseJSEnvironnement.ACCESS_PROTO) ?
                                                                                Object.getPrototypeOf(prev) :
                                                                                prev[curr];
                                                                 }, global);
                                          };

    // Exposing to the outside world
    global.traverseJSEnvironnement = traverseJSEnvironnement;

})(this);


